from PySide6.QtCore import QSize
from PySide6.QtWidgets import QLabel, QVBoxLayout

from pyside_app_core.qt.standard import MainWindow
from pyside_app_core.qt.widgets.connection_manager import ConnectionManager
from pyside_app_core.qt.widgets.core_icon import CoreIcon
from pyside_app_core.qt.widgets.multi_combo_box import MultiComboBox


class SimpleMainWindow(MainWindow):
    def __init__(self) -> None:
        super().__init__()

        # ------------------------------------------------------------------------------
        self.setMinimumSize(QSize(480, 240))

        _tool_bar = self.addToolBar("main")
        _tool_bar.setObjectName("main-tool-bar")
        plug_action = _tool_bar.addAction(
            CoreIcon(
                ":/core/iconoir/ev-plug-charging.svg",
                ":/core/iconoir/ev-plug-xmark.svg",
            ),
            "Connect",
        )
        plug_action.setCheckable(True)
        plug_action2 = _tool_bar.addAction(
            CoreIcon(
                ":/core/iconoir/ev-plug-charging.svg",
                ":/core/iconoir/ev-plug-xmark.svg",
            ),
            "Connect",
        )
        plug_action2.setCheckable(True)
        plug_action2.setChecked(True)
        reload_action = _tool_bar.addAction(
            CoreIcon(
                ":/core/iconoir/refresh-circle.svg",
            ),
            "Reload",
        )
        reload_action.setDisabled(True)
        _raise_action = _tool_bar.addAction(
            CoreIcon(
                ":/core/iconoir/floppy-disk.svg",
            ),
            "Save",
        )

        def _raise() -> None:
            raise Exception("This is a test error")  # noqa

        _raise_action.triggered.connect(_raise)

        # -----
        _central_layout = QVBoxLayout()
        self.centralWidget().setLayout(_central_layout)

        _heading = QLabel("Examples")
        _central_layout.addWidget(_heading)

        _multi_combo = MultiComboBox[str](placeholder_text="Options", parent=self)
        _multi_combo.addItems(["one", "two", "three", "four"])
        _central_layout.addWidget(_multi_combo)

        _con_mgr = ConnectionManager(parent=self)
        _central_layout.addWidget(_con_mgr)

        _central_layout.addStretch()

        self.statusBar().showMessage("Hi There")



from PySide6 import QtWidgets
from PySide6.QtWidgets import QSizePolicy

from pyside_app_core.qt.widgets.object_name_mixin import ObjectNameMixin


class DynamicStackedWidget(ObjectNameMixin, QtWidgets.QStackedWidget):
    def __init__(self, parent: QtWidgets.QWidget) -> None:
        super().__init__(parent=parent)

        self.setSizePolicy(QSizePolicy.Policy.Maximum, QSizePolicy.Policy.Maximum)
        self.setMinimumHeight(10)
        self.setMaximumHeight(100000)
        self.setContentsMargins(0, 0, 0, 0)



from PySide6.QtCore import QObject
from pytest_mock import MockerFixture

from pyside_app_core.services.serial_service.service import SerialService
from pyside_app_core.services.serial_service.types import (
    ChunkedData,
    Decodable,
    Encodable,
)


class MockCommand(Encodable, Decodable):
    def __init__(self, data: str):
        self._data = data

    def __repr__(self) -> str:
        return f"MockCommand<{self._data}>"

    def __str__(self) -> str:
        return repr(self)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, MockCommand):
            return self._data == other._data
        raise TypeError(f"Can't compare type: {type(other)} to {self.__class__.__name__}")

    def encode(self) -> bytes:
        return self._data.encode("utf-8")

    @classmethod
    def decode(cls, data: bytes) -> "MockCommand":
        return cls(data.decode("utf-8"))


class MockDecoder:
    @classmethod
    def process_buffer(cls, buffer: bytearray) -> ChunkedData:
        if b"\x00" in buffer:
            chunks: list[bytearray]
            remainder: bytearray
            *chunks, remainder = buffer.split(b"\x00")
            return chunks, remainder

        return [], None

    @classmethod
    def encode(cls, data: MockCommand) -> bytes:
        return data.encode()

    @classmethod
    def decode(cls, raw: bytearray) -> MockCommand:
        return MockCommand.decode(raw)


def test_serial_service_buffering(mocker: MockerFixture) -> None:
    """
    check that cors segmentation and buffering is proper.
    if we get partial messages we will buffer until the null bit terminator
    is read and emit only complete chunks.
    """
    svc = SerialService(transcoder=MockDecoder, parent=QObject())  # type: ignore[arg-type]

    mock_com = mocker.patch.object(svc, "_com")
    mock_data_sig = mocker.patch.object(svc, "data")

    # list of individual data reads from com port
    mock_com.readAll.side_effect = [
        b"abc\x00",
        b"123",
        b"xyz\x00987",
        b"654\x00abc\x00AAA",
    ]

    svc._on_data()
    mock_data_sig.emit.assert_called_once_with(MockCommand("abc"))
    assert mock_data_sig.emit.call_count == 1

    svc._on_data()
    assert svc._buffer == b"123"
    assert mock_data_sig.emit.call_count == 1

    svc._on_data()
    mock_data_sig.emit.assert_called_with(MockCommand("123xyz"))
    assert svc._buffer == b"987"
    assert mock_data_sig.emit.call_count == 2

    svc._on_data()
    mock_data_sig.emit.assert_any_call(MockCommand("987654"))
    mock_data_sig.emit.assert_called_with(MockCommand("abc"))
    assert svc._buffer == b"AAA"
    assert mock_data_sig.emit.call_count == 4



from PySide6.QtCore import QSize
from PySide6.QtGui import QIcon
from PySide6.QtWidgets import QPushButton, QWidget

from pyside_app_core.qt.widgets.object_name_mixin import ObjectNameMixin


class ToolButton(ObjectNameMixin, QPushButton):
    def __init__(self, icon: QIcon, tooltip: str, parent: QWidget):
        super().__init__(icon=icon, parent=parent)

        self.setMinimumSize(24, 24)

        self.setToolTip(tooltip)
        self.setContentsMargins(0, 0, 0, 0)

    def setFixedSize(  # type: ignore[override] # noqa
        self, size: QSize | tuple[int, int]
    ) -> None:
        if isinstance(size, tuple):
            size = QSize(size[0], size[1])
        super().setFixedSize(size + QSize(10, 10))
        self.setIconSize(size)



import subprocess
import tempfile
from pathlib import Path
from typing import Literal

from jinja2 import Environment, PackageLoader

from pyside_app_core.resource_generator.resource_types import (
    QtResourceFile,
    QtResourceGroup,
)

_std_resource_root = Path(__file__).parent.parent / "resources" / "core"


def _compile_qrc_template(resources: list[QtResourceGroup]) -> str:
    env = Environment(loader=PackageLoader("pyside_app_core.resources"), autoescape=True)
    qrc_template = env.get_template("resources.qrc.jinja2")
    return qrc_template.render(qresources=resources)


def _write_file(target: Path, data: str) -> Path:
    with open(target, "w") as f:
        f.write(data)

    return target


ResourceFormat = Literal["python", "binary"]


def _build_resource_groups(root_paths: list[Path]) -> list[QtResourceGroup]:
    groups: list[QtResourceGroup] = []
    root_prefixes: list[str] = []

    for root_path in root_paths:
        if not root_path.is_dir():
            continue

        prefix = root_path.name
        if prefix in root_prefixes:
            raise ValueError(f"Root path {root_path} has duplicate prefix {prefix}")

        root_prefixes.append(prefix)

        files: list[QtResourceFile] = []

        for item in root_path.glob("**/*"):
            if item.is_dir():
                continue
            files.append(
                QtResourceFile(
                    str(item),
                    item.relative_to(root_path).as_posix(),
                )
            )

        groups.append(
            QtResourceGroup(
                prefix,
                None,
                files,
            )
        )

    return groups


def compile_qrc_to_resources(
    target_dir: Path,
    rcc_format: ResourceFormat = "binary",
    additional_resource_roots: list[Path] | None = None,
    *,
    debug: bool = False,
) -> Path:
    with tempfile.TemporaryDirectory() as tempdir:
        temp_path = Path(tempdir)

        resources = _build_resource_groups(
            [
                _std_resource_root,
                *(additional_resource_roots or []),
            ]
        )

        qrc_file = _write_file(temp_path / "resources.qrc", _compile_qrc_template(resources))

        rcc_args = []

        if rcc_format == "binary":
            file_name = "resources.rcc"
            rcc_args.append("--binary")
        else:
            file_name = "resources.py"

        file_target = target_dir / file_name

        subprocess.check_call(["pyside6-rcc", *rcc_args, "-o", str(file_target), str(qrc_file)])

        if debug:
            print("-" * 80)
            print(qrc_file.read_text())
            print("-" * 80)

    return file_target



from typing import cast

from PySide6.QtCore import QRect, QSize
from PySide6.QtGui import (
    QColorConstants,
    QGuiApplication,
    QIcon,
    QIconEngine,
    QPainter,
    QPalette,
    QPixmap,
)
from PySide6.QtSvg import QSvgRenderer


class _SvgIconEngine(QIconEngine):
    def __init__(self, on_state: str, off_state: str | None):
        super().__init__()
        self._on_file = on_state
        self._off_file = off_state

    def clone(self) -> "_SvgIconEngine":
        return self.__class__(self._on_file, self._off_file)

    def pixmap(self, size: QSize, mode: QIcon.Mode, state: QIcon.State) -> QPixmap:
        pixmap = QPixmap(size)
        pixmap.fill(QColorConstants.Transparent)

        painter = QPainter(pixmap)
        self.paint(painter, pixmap.rect(), mode, state)
        painter.end()

        return pixmap

    def paint(self, painter: QPainter, rect: QRect, mode: QIcon.Mode, state: QIcon.State) -> None:
        app_palette = cast(QGuiApplication, QGuiApplication.instance()).palette()

        renderer = (
            QSvgRenderer(self._on_file) if state == QIcon.State.On else QSvgRenderer(self._off_file or self._on_file)
        )

        if mode == QIcon.Mode.Normal:
            color = app_palette.color(QPalette.ColorRole.Text)
        elif mode == QIcon.Mode.Disabled:
            color = app_palette.color(QPalette.ColorRole.PlaceholderText)
        elif mode == QIcon.Mode.Active:
            color = app_palette.color(QPalette.ColorRole.Text)
        elif mode == QIcon.Mode.Selected:
            color = app_palette.color(QPalette.ColorRole.BrightText)
        else:
            color = app_palette.color(QPalette.ColorRole.Text)

        painter.save()
        renderer.render(painter)
        painter.setCompositionMode(QPainter.CompositionMode.CompositionMode_SourceIn)
        painter.fillRect(rect, color)
        painter.restore()


class CoreIcon(QIcon):
    def __init__(self, on_state: str, off_state: str | None = None):
        super().__init__(_SvgIconEngine(on_state, off_state))



import contextlib
from collections.abc import Iterator
from typing import Literal

from PySide6.QtCore import QSize, Qt
from PySide6.QtGui import QAction, QIcon
from PySide6.QtWidgets import QMainWindow, QToolBar, QToolButton

from pyside_app_core.qt.widgets.object_name_mixin import ObjectNameMixin

ToolBarArea = Literal["top", "bottom", "left", "right"]

_TOOL_BAR_AREA_MAP = {
    "top": Qt.ToolBarArea.TopToolBarArea,
    "bottom": Qt.ToolBarArea.BottomToolBarArea,
    "right": Qt.ToolBarArea.RightToolBarArea,
    "left": Qt.ToolBarArea.LeftToolBarArea,
}


class ToolBarContext(ObjectNameMixin, QToolBar):
    def __init__(self, area: ToolBarArea, parent: QMainWindow, *, movable: bool = False) -> None:
        self._area = area
        self._actions: list[QAction] = []

        if area == "left":
            self._border_side = "right"
        elif area == "right":
            self._border_side = "left"
        elif area == "top":
            self._border_side = "bottom"
        else:
            self._border_side = "top"

        self.OBJECT_NAME = f"ToolBar_{area}"
        super().__init__(self.OBJECT_NAME, parent=parent)

        self.setContentsMargins(0, 0, 0, 0)
        self.setGeometry(10, 10, self.width(), self.height())
        self.setMovable(movable)
        self.setIconSize(QSize(28, 28))

        parent.addToolBar(_TOOL_BAR_AREA_MAP[self._area], self)

    @contextlib.contextmanager
    def add_action(self, name: str, icon: QIcon | None = None) -> Iterator[QAction]:
        action = QAction(text=name, parent=self)
        action.setObjectName(f"ToolBarAction_{name}")
        if icon:
            action.setIcon(icon)

        self._actions.append(action)
        if len(self._actions) == 1:
            _ = next(c for c in self.children() if isinstance(c, QToolButton))
        yield action
        self.addAction(action)



def wrap_text(text: str, *, width: int = 80, padded: bool = False) -> str:
    lines = text.splitlines()

    for i, line in enumerate(lines):
        if len(line) > width:
            head, tail = line[:width], line[width:]
            lines[i] = head
            lines.insert(i + 1, tail)

    if padded:
        lines = [f"{ln: <{width}}" for ln in lines]

    return "\n".join(lines)



import argparse
import os
from pathlib import Path

from pyside_app_core import log

_root = Path(__file__).parent


def compile_rcc() -> None:
    from pyside_app_core import resource_generator

    log.info("Generating resources...")

    resource_generator.compile_qrc_to_resources(target_dir=_root, debug=True)


def main() -> None:
    # NOTE: these imports are lumped in here for the ability to
    # set PLATFORM_OVERRIDE from code, normally you would not need
    # to do this
    from pyside_app_core.errors import excepthook
    from pyside_app_core.qt.standard.error_dialog import ErrorDialog
    from toolbar_app import __version__
    from toolbar_app.app import SimpleApp

    excepthook.install_excepthook(ErrorDialog)
    from pyside_app_core.qt.application_service import AppMetadata

    AppMetadata.init(
        "com.example.simple-app",
        "Simple App",
        __version__,
        help_url="https://github.com/leocov-dev/pyside-app-core",
        bug_report_url="https://github.com/leocov-dev/pyside-app-core/issues",
    )

    SimpleApp().launch()


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--style", choices=["mac", "win", "nix"])
    parser.add_argument("--generate-rcc", action="store_true")

    args = parser.parse_args()

    match args.style:
        case "mac":
            os.environ["PLATFORM_OVERRIDE"] = "Darwin"
        case "win":
            os.environ["PLATFORM_OVERRIDE"] = "Windows"
        case "nix":
            os.environ["PLATFORM_OVERRIDE"] = "Linux"

    if args.generate_rcc:
        compile_rcc()

    main()



class CoreError(Exception):
    """base error for application"""

    @property
    def internal(self) -> bool:
        return self._internal

    def __init__(self, msg: str, *, internal: bool = False):
        self._internal = internal
        super().__init__(msg)


class ApplicationError(CoreError):
    """something bad happened at the application level"""



from typing import cast

from PySide6.QtCore import QCoreApplication, QObject, QSettings
from PySide6.QtGui import QShowEvent
from PySide6.QtWidgets import QWidget

from pyside_app_core.qt.application_service import AppMetadata


class SettingsMixin:
    def __init__(self, parent: QObject | None = None, *args: object, **kwargs: object):
        super().__init__(*args, **kwargs)

        self._settings = QSettings(
            AppMetadata.id,
            AppMetadata.name,
            parent,
        )

        self._restored = False

        cast(QCoreApplication, QCoreApplication.instance()).aboutToQuit.connect(self._store_state)

    def showEvent(self, event: QShowEvent) -> None:  # noqa: N802
        if not self._restored:
            self._restore_state()
            self._restored = True

        if isinstance(self, QWidget):
            cast(QWidget, super()).showEvent(event)

    def _restore_state(self) -> None:
        pass

    def _store_state(self) -> None:
        pass



from pyside_app_core.constants import LIST_DATA_LEN_MAX
from pyside_app_core.errors.basic_errors import CoreError


class EncodingListError(CoreError):
    def __init__(self, actual_len: int):
        super().__init__(
            f"command data list was too long, max: {LIST_DATA_LEN_MAX}, actual: {actual_len}",
            internal=True,
        )



import argparse
from pathlib import Path

from pyside_app_core.resource_generator import compile_qrc_to_resources


def main() -> None:
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "-r",
        "--extra-resource-root",
        help="Provide a root path to search for additional QResource files " "(Can specify this flag multiple times)",
        action="append",
    )

    parser.add_argument(
        "target_dir",
        help="Provide a directory path were the final resource file will be placed",
        metavar="target-dir",
    )

    parser.add_argument(
        "--debug",
        help="print the contents of the qrc file to stdout",
        action="store_true",
    )

    args = parser.parse_args()

    print("...")

    generated_file = compile_qrc_to_resources(
        target_dir=Path(args.target_dir),
        additional_resource_roots=[Path(r) for r in args.extra_resource_root or []],
        debug=args.debug,
    )

    print(f'Generated: "{generated_file}"')


if __name__ == "__main__":
    main()



from PySide6.QtCore import Qt, QUrl
from PySide6.QtGui import QAction, QDesktopServices
from PySide6.QtWidgets import QMainWindow, QWidget

from pyside_app_core.qt.application_service import AppMetadata
from pyside_app_core.qt.standard.about_dialog import AboutDialog
from pyside_app_core.qt.standard.base_window import BaseMixin
from pyside_app_core.qt.widgets.menu_ctx import MenuBarContext
from pyside_app_core.qt.widgets.tool_bar_ctx import ToolBarContext
from pyside_app_core.qt.widgets.window_settings_mixin import WindowSettingsMixin
from pyside_app_core.services import platform_service


class MainWindow(WindowSettingsMixin, BaseMixin, QMainWindow):
    def __init__(self) -> None:
        super().__init__(parent=None)

        self._about_dialog = AboutDialog()

        self._central = QWidget(parent=self)
        self.setCentralWidget(self._central)

        # must call in order to show grab handle
        self.statusBar().show()

        self._menu_bar = MenuBarContext(self)
        self._menu_bar.setNativeMenuBar(platform_service.is_macos)
        self.setMenuBar(self._menu_bar)

        with (
            self._menu_bar.menu("File") as file_menu,
            file_menu.action("Quit") as exit_action,
        ):
            exit_action.setMenuRole(QAction.MenuRole.QuitRole)
            exit_action.triggered.connect(self.close)

        self._build_menus()

        with self._menu_bar.menu("Window") as window_menu:
            with window_menu.action("Minimize") as min_action:
                min_action.setShortcut("Ctrl+M")
                min_action.triggered.connect(self.showMinimized)
            with window_menu.action("Zoom") as zoom_action:
                zoom_action.triggered.connect(self.showMaximized)

        with self._menu_bar.menu("Help") as help_menu:
            with help_menu.action("About") as about_action:
                about_action.setMenuRole(QAction.MenuRole.AboutRole)
                about_action.triggered.connect(self._about_dialog.exec)

            if AppMetadata.help_url:
                with help_menu.action("Get Help") as help_action:
                    help_action.triggered.connect(lambda: QDesktopServices.openUrl(QUrl(AppMetadata.help_url)))

            if AppMetadata.bug_report_url:
                with help_menu.action("Report Bug") as help_action:
                    help_action.triggered.connect(lambda: QDesktopServices.openUrl(QUrl(AppMetadata.bug_report_url)))

    @property
    def menu_bar(self) -> MenuBarContext:
        return self._menu_bar

    def _build_menus(self) -> None:
        pass


class MainToolbarWindow(MainWindow):
    def __init__(self) -> None:
        super().__init__()

        self._tool_bar = ToolBarContext(area="top", parent=self, movable=False)
        self._tool_bar.setToolButtonStyle(Qt.ToolButtonStyle.ToolButtonIconOnly)

    @property
    def tool_bar(self) -> ToolBarContext:
        return self._tool_bar



import pprint
import sys
from pathlib import Path

from PySide6.QtCore import QResource

from pyside_app_core import log


def assert_resources_file(rcc: Path | None) -> Path:
    """if the given rcc file is None or not found then will try to get a 'default' resource file"""
    if rcc and rcc.exists():
        return rcc

    from pyside_app_core.services import debug_service

    tried = []

    for i in range(10):
        try:
            caller = debug_service.get_caller_file(2, i)
            rcc = caller.parent / "resources.rcc"
            tried.append(str(rcc))
            if rcc.exists():
                break
        except IndexError:
            # ignore caller frame depth index errors
            pass
    else:
        log.error(
            f"No resource.rcc file given or found, attempted:\n"
            f'{pprint.pformat([t for t in sorted(set(tried)) if "/" in t], compact=False)}\n'
            f"Will now exit.",
            file=sys.stderr,
        )
        sys.exit(1)

    return rcc


def register_resource_file(rcc: Path | None) -> None:
    # registering a binary rcc is marginally faster than importing a python resource module
    QResource.registerResource(str(assert_resources_file(rcc)))



from collections.abc import Sequence
from typing import Protocol

from PySide6.QtCore import Slot
from PySide6.QtSerialPort import QSerialPort, QSerialPortInfo


class Encodable(Protocol):
    def encode(self) -> bytes: ...

    def __str__(self) -> str: ...


class Decodable(Protocol):
    @classmethod
    def decode(cls, data: bytes) -> "Decodable": ...

    def __str__(self) -> str: ...


ChunkedData = tuple[Sequence[bytearray], bytearray | None]


class TranscoderInterface(Protocol):
    @classmethod
    def process_buffer(cls, buffer: bytearray) -> ChunkedData: ...

    @classmethod
    def encode(cls, data: Encodable) -> bytes: ...

    @classmethod
    def decode(cls, raw: bytearray) -> Decodable: ...


class SerialReader(Protocol):
    @Slot()
    def handle_serial_connect(self, com: QSerialPort) -> None: ...

    @Slot()
    def handle_serial_disconnect(self) -> None: ...

    @Slot()
    def handle_serial_ports(self, ports: list[QSerialPortInfo]) -> None: ...

    @Slot()
    def handle_serial_data(self, data: Decodable) -> None: ...

    @Slot()
    def handle_serial_error(self, error: Exception) -> None: ...


class PortFilter(Protocol):
    def __call__(self, ports: list[QSerialPortInfo]) -> list[QSerialPortInfo]: ...



import struct
from collections.abc import Iterator, Mapping
from decimal import Decimal
from typing import Any, TypeVar

from pyside_app_core.constants import (
    DATA_STRUCT_ENDIAN,
    FLOAT_PRECISION,
    STRUCT_FLOAT_FMT,
)
from pyside_app_core.services.serial_service import conversion_utils
from pyside_app_core.utils import compare

K = TypeVar("K", bound=int)

_TWO_BYTE_LEN = 65535


class FloatMap(Mapping[K, float]):
    _count_fmt = "H"  # unsigned short, 2 bytes
    _key_fmt = "H"  # unsigned short, 2 bytes

    def __init__(self, data: Mapping[K, float]):
        self._data = data

        if not data:
            raise ValueError(f"can't create an empty {self.__class__.__name__}")

        if len(data) > _TWO_BYTE_LEN:
            raise ValueError("data has too many values, must fit in 2 bytes")

        for k in data:
            if k > _TWO_BYTE_LEN:
                raise ValueError(f'key: "{k}" does not fit in 2 bytes')

            if k < 0:
                raise ValueError(f'key: "{k}" is not a positive number')

    def __getitem__(self, k: K) -> float:
        return self._data[k]

    def __len__(self) -> int:
        return len(self._data)

    def __iter__(self) -> Iterator[K]:
        return iter(self._data)

    def __repr__(self) -> str:
        return self._data.__repr__()

    def __str__(self) -> str:
        return self._data.__str__()

    def __eq__(self, other: "FloatMap[K]") -> bool:  # type: ignore[override]
        val_eq: list[bool] = []
        for k, v in self.items():
            other_v = other.get(k)
            if other_v:
                val_eq.append(compare.float_approx(v, other_v))

        return all(
            [
                sorted(self.keys()) == sorted(other.keys()),
                *val_eq,
            ]
        )

    @classmethod
    def _key_xform(cls, key: int) -> Any:
        return key

    @classmethod
    def pack_format(cls, item_count: int) -> str:
        if item_count < 1:
            raise ValueError("must pack at least 1 item")

        accum_fmt = f"{DATA_STRUCT_ENDIAN}{cls._count_fmt}"

        accum_fmt += f"{cls._key_fmt}{STRUCT_FLOAT_FMT}" * item_count

        return accum_fmt

    def pack(self) -> bytearray:
        """
        pattern:
        count,key,value,key,value,key,value,...
        H - count, number of pairs - max 2 bytes unsigned...
        H - key in 2 bytes corresponding to enum ConfigurationParam
        f/d - value in 4 or 8 bytes corresponding to float/double value

        example:
        struct.pack("<HHfHf", count, key, value, key, value)
        """
        if not self._data:
            raise AttributeError("Command had no data")

        count = len(self)
        flattened_data = []

        for key, value in self._data.items():
            flattened_data.extend([key, value])

        raw = struct.pack(self.pack_format(count), count, *flattened_data)
        return bytearray(raw)

    @classmethod
    def unpack(cls, raw_data: bytes) -> "FloatMap[K]":
        count_bytes = 2
        raw_pair_count = raw_data[:count_bytes]
        raw_key_val = raw_data[count_bytes:]

        pair_count = conversion_utils.int_from_bytes(raw_pair_count, signed=False)
        fmt_chars = [cls._key_fmt, STRUCT_FLOAT_FMT] * pair_count

        flat_pairs = struct.unpack(f"{DATA_STRUCT_ENDIAN}{''.join(fmt_chars)}", raw_key_val)

        data = {}
        iterable = iter(flat_pairs)

        for key in iterable:
            val = next(iterable)
            if FLOAT_PRECISION == "single":
                val = float(Decimal(val).quantize(Decimal("0.000000")))
            data[cls._key_xform(key)] = val

        return cls(data=data)



import inspect
import logging
from collections.abc import Callable
from typing import Protocol

__logger_cache: dict[str, logging.Logger] = {}

__pac_name = "pyside_app_core"


def _get_caller_name() -> str:
    try:
        stack = inspect.stack()
        frame = stack[2][0]
        module = inspect.getmodule(frame)
        if not module:
            return __pac_name
    except:  # noqa
        return __pac_name
    name = module.__name__
    if name == "__main__":
        name = __pac_name

    return name


def _get_cached_logger(name: str) -> logging.Logger:
    if name in __logger_cache:
        lg = __logger_cache[name]
    else:
        lg = logging.getLogger(name)
        __logger_cache[name] = lg

    return lg


class _Logger(Protocol):
    def debug(self, msg: str, /, *args: object, **kwargs: object) -> None: ...

    def info(self, msg: str, /, *args: object, **kwargs: object) -> None: ...

    def warning(self, msg: str, /, *args: object, **kwargs: object) -> None: ...

    def error(self, msg: str, /, *args: object, **kwargs: object) -> None: ...

    def critical(self, msg: str, /, *args: object, **kwargs: object) -> None: ...

    def exception(self, msg: str, /, *args: object, **kwargs: object) -> None: ...


def __default_get_logger() -> logging.Logger:
    name = _get_caller_name()
    return _get_cached_logger(name)


GetLogger = Callable[[], _Logger]

__get_logger_func: GetLogger = __default_get_logger  # type: ignore[assignment]


def configure_get_logger_func(func: GetLogger) -> None:
    global __get_logger_func  # noqa: PLW0603
    __get_logger_func = func


def debug(msg: object, *args: object, **kwargs: object) -> None:
    """call logger's debug"""
    lg = __get_logger_func()
    lg.debug(msg, *args, **kwargs)  # type: ignore[arg-type]


def info(msg: object, *args: object, **kwargs: object) -> None:
    """call logger's info"""
    lg = __get_logger_func()
    lg.info(msg, *args, **kwargs)  # type: ignore[arg-type]


def warning(msg: object, *args: object, **kwargs: object) -> None:
    """call logger's warning"""
    lg = __get_logger_func()
    lg.warning(msg, *args, **kwargs)  # type: ignore[arg-type]


def error(msg: object, *args: object, **kwargs: object) -> None:
    """call logger's error"""
    lg = __get_logger_func()
    lg.error(msg, *args, **kwargs)  # type: ignore[arg-type]


def critical(msg: object, *args: object, **kwargs: object) -> None:
    """call logger's critical"""
    lg = __get_logger_func()
    lg.critical(msg, *args, **kwargs)  # type: ignore[arg-type]


def exception(msg: object, *args: object, **kwargs: object) -> None:
    """call logger's exception"""
    lg = __get_logger_func()
    lg.exception(msg, *args, **kwargs)  # type: ignore[arg-type]



import inspect
from pathlib import Path


def get_caller_file(frame: int = 2, depth: int = 3) -> Path:
    # TODO: not sure this will reliably find the entrypoint
    cur_frame = inspect.currentframe()
    call_frame = inspect.getouterframes(cur_frame, frame)
    return Path(call_frame[depth].filename)



"""
Adapted from:
https://gis.stackexchange.com/questions/350148/qcombobox-multiple-selection-pyqt5
"""

from collections.abc import Generator
from typing import Generic, TypeVar, cast

from PySide6.QtCore import (
    QEvent,
    QModelIndex,
    QObject,
    QPersistentModelIndex,
    QSize,
    Qt,
    QTimerEvent,
    Signal,
)
from PySide6.QtGui import (
    QFontMetrics,
    QMouseEvent,
    QResizeEvent,
    QStandardItem,
    QStandardItemModel,
)
from PySide6.QtWidgets import (
    QComboBox,
    QStyledItemDelegate,
    QStyleOptionViewItem,
    QWidget,
)

from pyside_app_core.qt.widgets.object_name_mixin import ObjectNameMixin
from pyside_app_core.qt.widgets.settings_mixin import SettingsMixin

DT = TypeVar("DT")


class MultiComboBox(ObjectNameMixin, SettingsMixin, QComboBox, Generic[DT]):
    """a combo box of checkable items.
    The selected options will be listed as text in the combo box line.
    The first option is a show/hide all option.
    """

    class Delegate(QStyledItemDelegate):
        def sizeHint(  # noqa:N802
            self,
            option: QStyleOptionViewItem,
            index: QModelIndex | QPersistentModelIndex,
        ) -> QSize:
            size = super().sizeHint(option, index)
            size.setHeight(20)
            return size

    selectionChanged = Signal(list)  # noqa:N815

    def __init__(
        self,
        placeholder_text: str,
        all_text: str = "Show/Hide All",
        parent: QWidget | None = None,
        *args: object,
        **kwargs: object,
    ) -> None:
        super().__init__(*args, parent=parent, **kwargs)

        # Make the combo editable to set a custom text, but readonly
        self.setEditable(True)
        self.lineEdit().setReadOnly(True)
        self.lineEdit().setPlaceholderText(placeholder_text)

        # Use custom delegate
        self.setItemDelegate(MultiComboBox.Delegate())

        # Hide and show popup when clicking the line edit
        self.lineEdit().installEventFilter(self)
        self.closeOnLineEditClick = False

        # Prevent popup from closing when clicking on an item
        self.view().viewport().installEventFilter(self)

        self.addItem(all_text, is_checked=True)

        self.model().itemChanged.connect(self._update_selection)
        self.model().dataChanged.connect(self._update_text)
        self.model().dataChanged.connect(self._emit_current_data)

    def model(self) -> QStandardItemModel:
        return cast(QStandardItemModel, super().model())

    def _emit_current_data(self) -> None:
        self.selectionChanged.emit(self.currentData())

    def _update_selection(self, item: QStandardItem) -> None:
        if item.index().row() == 0:
            # update on the show/hide all case
            state = item.checkState()
            self.model().layoutAboutToBeChanged.emit()
            for item in self.modelItems():
                item.setCheckState(state)
            self.model().layoutChanged.emit()

    def resizeEvent(self, event: QResizeEvent) -> None:  # noqa: N802
        # Recompute text to elide as needed
        self._update_text()
        super().resizeEvent(event)

    def eventFilter(self, watched: QObject, event: QEvent) -> bool:  # noqa: N802
        if watched == self.lineEdit():
            if event.type() == QEvent.Type.MouseButtonRelease:
                if self.closeOnLineEditClick:
                    self.hidePopup()
                else:
                    self.showPopup()
                return True
            return False

        if watched == self.view().viewport() and (
            isinstance(event, QMouseEvent) and event.type() == QEvent.Type.MouseButtonRelease
        ):
            index = self.view().indexAt(event.pos())
            item = self.model().item(index.row())

            if item.checkState() == Qt.CheckState.Checked:
                item.setCheckState(Qt.CheckState.Unchecked)
            else:
                item.setCheckState(Qt.CheckState.Checked)
            return True
        return False

    def showPopup(self) -> None:  # noqa: N802
        super().showPopup()
        # When the popup is displayed, a click on the lineedit should close it
        self.closeOnLineEditClick = True

    def hidePopup(self) -> None:  # noqa: N802
        super().hidePopup()
        # Used to prevent immediate reopening when clicking on the lineEdit
        self.startTimer(100)
        # Refresh the display text when closing
        self._update_text()

    def timerEvent(self, event: QTimerEvent) -> None:  # noqa: N802
        # After timeout, kill timer, and re-enable click on line edit
        self.killTimer(event.timerId())
        self.closeOnLineEditClick = False

    def _update_text(self) -> None:
        texts = [i.text() for i in self.modelItems() if i.checkState() == Qt.CheckState.Checked]
        text = ", ".join(texts)

        # Compute elided text (with "...")
        metrics = QFontMetrics(self.lineEdit().font())
        text = metrics.elidedText(text, Qt.TextElideMode.ElideRight, self.lineEdit().width())
        self.lineEdit().setText(text)

    def addItem(  # type:ignore[override] # noqa
        self,
        text: str,
        user_data: DT | None = None,
        *,
        is_checked: bool = False,
    ) -> None:
        item = QStandardItem()
        item.setText(text)
        if user_data is None:
            item.setData(text, Qt.ItemDataRole.UserRole)
        else:
            item.setData(user_data, Qt.ItemDataRole.UserRole)

        item.setFlags(Qt.ItemFlag.ItemIsEnabled | Qt.ItemFlag.ItemIsUserCheckable)
        item.setCheckState(Qt.CheckState.Checked if is_checked else Qt.CheckState.Unchecked)

        self.model().appendRow(item)

    def addItems(  # type:ignore[override] # noqa
        self,
        texts: list[str],
        data_list: list[DT] | None = None,
        *,
        all_checked: bool = False,
    ) -> None:
        for i, text in enumerate(texts):
            try:
                data = (data_list or [])[i]
            except (TypeError, IndexError):
                data = None
            self.addItem(text, data, is_checked=all_checked)

    def currentData(  # type: ignore[override] # noqa
        self,
        role: Qt.ItemDataRole = Qt.ItemDataRole.UserRole,
    ) -> list[DT]:
        return [i.data(role) for i in self.modelItems() if i.checkState() == Qt.CheckState.Checked]

    def currentOptions(self) -> list[tuple[str, DT]]:  # noqa: N802
        return [
            (i.text(), i.data(Qt.ItemDataRole.UserRole))
            for i in self.modelItems()
            if i.checkState() == Qt.CheckState.Checked
        ]

    def modelItems(self) -> Generator[QStandardItem, None, None]:  # noqa: N802
        for i in range(1, self.model().rowCount()):
            yield self.model().item(i)

    def _restore_state(self) -> None:
        options = cast(list[str], self._settings.value(f"{self.obj_name}_options"))
        if not options:
            return

        for item in self.modelItems():
            state = Qt.CheckState.Checked if item.text() in options else Qt.CheckState.Unchecked
            item.setCheckState(state)

    def _store_state(self) -> None:
        self._settings.setValue(f"{self.obj_name}_options", [n for n, _ in self.currentOptions()])



import sys
from pathlib import Path
from typing import Generic, TypeVar

from PySide6.QtCore import Qt
from PySide6.QtGui import QIcon
from PySide6.QtWidgets import QApplication, QMainWindow, QStyle

from pyside_app_core import log
from pyside_app_core.errors.basic_errors import ApplicationError
from pyside_app_core.qt import register_resource_file
from pyside_app_core.qt.application_service import AppMetadata

M = TypeVar("M", bound=QMainWindow)


class BaseApp(QApplication, Generic[M]):
    def __init__(self, resources_rcc: Path | None = None, *args: object, **kwargs: object) -> None:
        super().__init__(*args, **kwargs)
        self.setStyle("fusion")
        self.setAttribute(Qt.ApplicationAttribute.AA_UseStyleSheetPropagationInWidgetStyles)

        register_resource_file(resources_rcc)

        self.setWindowIcon(
            QIcon(AppMetadata.icon)
            if AppMetadata.icon
            else self.style().standardIcon(QStyle.StandardPixmap.SP_FileDialogDetailedView)
        )

        self._main_window: M = self.build_main_window()

    def build_main_window(self) -> M:
        raise NotImplementedError

    def launch(self) -> None:
        if not self._main_window:
            raise ApplicationError(f"Must subclass {BaseApp.__name__} and define a main window")

        self._main_window.show()

        try:
            self.about_to_exit(self.exec())
        except Exception as e:  # noqa:BLE001
            log.exception(e)
            sys.exit(1)

    def about_to_exit(self, ret_code: int) -> int:
        return ret_code



from typing import cast

from PySide6.QtCore import QSize, Qt, QTimer, Signal, Slot
from PySide6.QtSerialPort import QSerialPort, QSerialPortInfo
from PySide6.QtWidgets import QComboBox, QHBoxLayout, QPushButton, QVBoxLayout, QWidget

from pyside_app_core import log
from pyside_app_core.qt.widgets.core_icon import CoreIcon
from pyside_app_core.qt.widgets.settings_mixin import SettingsMixin
from pyside_app_core.utils.time_ms import SECONDS


class ConnectionManager(SettingsMixin, QWidget):
    refresh_ports = Signal()
    request_connect = Signal(QSerialPortInfo)
    request_disconnect = Signal()

    SIZE = QSize(30, 30)
    ICON_SIZE = QSize(20, 20)

    def __init__(
        self,
        *,
        remember_last_connection: bool = False,
        parent: QWidget | None = None,
    ):
        super().__init__(parent=parent)

        self._remember_last_connection = remember_last_connection

        _ly = QVBoxLayout()
        _ly.setSpacing(0)
        _ly.setContentsMargins(5, 5, 5, 5)
        self.setLayout(_ly)

        _ly_list = QHBoxLayout()
        _ly_list.setSpacing(5)
        _ly.addLayout(_ly_list)

        _ly_actions = QHBoxLayout()
        _ly.addLayout(_ly_actions)

        # -------
        # refresh button
        self._refresh_btn = QPushButton(
            icon=CoreIcon(":/core/iconoir/refresh-circle.svg"),
            text="",
            parent=self,
        )
        self._refresh_btn.setFixedSize(self.SIZE)
        self._refresh_btn.setIconSize(self.ICON_SIZE)
        self._refresh_btn.setContentsMargins(0, 0, 0, 0)
        self._refresh_btn.setToolTip("Refresh Device List")
        _ly_list.addWidget(self._refresh_btn)

        # -------
        # port selection list
        self._port_list = QComboBox(self)
        self._port_list.setPlaceholderText("Choose A Device")
        self._port_list.setFixedHeight(self.SIZE.height())
        _ly_list.addWidget(self._port_list, stretch=9)

        # --------
        # connect/disconnect button
        self._connect_btn = QPushButton(
            icon=CoreIcon(
                ":/core/iconoir/ev-plug-charging.svg",
                ":/core/iconoir/ev-plug-xmark.svg",
            ),
            text="",
            parent=self,
        )
        self._connect_btn.setCheckable(True)
        self._connect_btn.setFixedSize(self.SIZE)
        self._connect_btn.setIconSize(self.ICON_SIZE)
        self._connect_btn.setContentsMargins(0, 0, 0, 0)
        self._connect_btn.setToolTip("Refresh Device List")
        _ly_list.addWidget(self._connect_btn)

        # ------------------------------------------------------------------------------
        # signals
        self._refresh_btn.clicked.connect(self.request_port_refresh)
        self._port_list.currentIndexChanged.connect(self._on_port_selected)

    @property
    def current_port(self) -> QSerialPortInfo | None:
        return cast(
            QSerialPortInfo | None,
            self._port_list.currentData(Qt.ItemDataRole.UserRole),
        )

    @Slot()
    def handle_serial_connect(self, com: QSerialPort) -> None:
        log.debug(f"Connected to serial port {com.portName()}")

    @Slot()
    def handle_serial_disconnect(self) -> None:
        log.debug("Com port disconnected")
        self.request_port_refresh()

    @Slot()
    def handle_serial_ports(self, ports: list[QSerialPortInfo]) -> None:
        self.setEnabled(True)
        self._port_list.clear()

        for port in ports:
            name = port.portName()
            log.debug(f"Adding port {name}")

            self._port_list.addItem(name, port)

    @Slot()
    def handle_serial_data(self, data: object) -> None:
        log.debug(f"Received data: {data}")

    @Slot()
    def handle_serial_error(self, error: Exception) -> None:
        pass

    def request_port_refresh(self) -> None:
        log.debug("Requesting Port Refresh")
        self.setDisabled(True)
        QTimer.singleShot(1 * SECONDS, self.refresh_ports.emit)

    def _on_port_selected(self, index: int) -> None:
        port = self._port_list.itemData(index, Qt.ItemDataRole.UserRole)
        log.debug(f"Port Selected: {port}")
        self.request_connect.emit(port)

    def _format_port_name(self, port: QSerialPortInfo) -> str:
        name: str = port.portName()
        extra = ""
        if mfc := port.manufacturer():
            prod_id = port.productIdentifier()
            extra = f"{mfc}/{prod_id}" if prod_id else mfc

        return f"{name} ({extra})" if extra else name

    def _store_state(self) -> None:
        if not self._remember_last_connection:
            return

        if self.current_port and self.current_port.serialNumber():
            self._settings.setValue("last_port_serial", self.current_port.serialNumber())

    def _restore_state(self) -> None:
        if not self._remember_last_connection:
            return None



from typing import Literal

from PySide6.QtCore import Slot
from PySide6.QtGui import QAction, QIcon
from PySide6.QtWidgets import QButtonGroup, QHBoxLayout, QMenu, QVBoxLayout, QWidget

from pyside_app_core.qt.widgets.dynamic_stacked_widget import DynamicStackedWidget
from pyside_app_core.qt.widgets.object_name_mixin import ObjectNameMixin
from pyside_app_core.qt.widgets.settings_mixin import SettingsMixin
from pyside_app_core.qt.widgets.tool_button import ToolButton

ToolStackSide = Literal["right", "left"]


class ToolStack(SettingsMixin, ObjectNameMixin, QWidget):
    def __init__(
        self,
        side: ToolStackSide,
        parent: QWidget,
        menu: QMenu | None = None,
    ):
        super().__init__(parent=parent)

        # opposite_side = "left" if side == "right" else "right"

        self._menu = menu

        container_layout = QHBoxLayout()
        container_layout.setContentsMargins(0, 0, 0, 0)
        container_layout.setSpacing(0)
        self.setLayout(container_layout)

        _button_container = QWidget(self)
        _button_container.setObjectName(f"{self.obj_name}_BUTTON_CONTAINER")

        self._button_layout = QVBoxLayout()
        _button_container.setLayout(self._button_layout)
        self._button_layout.setSpacing(8)

        self._button_group = QButtonGroup(self)
        self._button_group.setExclusive(False)

        self._stack = DynamicStackedWidget(self)

        if side == "right":
            self._button_layout.setContentsMargins(10, 10, 10, 3)
            container_layout.addWidget(self._stack)
            container_layout.addWidget(_button_container)
        else:
            self._button_layout.setContentsMargins(10, 10, 10, 3)
            container_layout.addWidget(_button_container)
            container_layout.addWidget(self._stack)

    def add_widget(self, icon: QIcon, widget: QWidget, tooltip: str) -> None:
        container = QWidget(self)
        container.setObjectName(f"{self.obj_name}_CONTAINER")
        container_layout = QVBoxLayout()
        container.setLayout(container_layout)
        container_layout.addWidget(widget)
        container_layout.addStretch()

        index: int = self._stack.addWidget(container)

        button = ToolButton(icon, tooltip, parent=self)
        button.setFixedSize((24, 24))
        button.setCheckable(True)
        self._button_group.addButton(button, index)

        num = self._button_layout.count()

        if num > 0:
            self._button_layout.insertWidget(num - 1, button)
        else:
            button.setChecked(True)
            self._button_layout.addWidget(button)
            self._button_layout.addStretch()

        button.clicked.connect(lambda checked: self._on_click(index, checked))

        if self._menu:
            action = QAction(text=tooltip, parent=self._menu)
            action.triggered.connect(lambda: self._on_menu_click(index))
            self._menu.addAction(action)

    @Slot(int, bool)
    def _on_click(self, index: int, checked: bool) -> None:  # noqa: FBT001
        for i, button in enumerate(self._button_group.buttons()):
            self._stack.setCurrentIndex(index)

            if index == i:
                self._stack.setVisible(checked)
                button.setChecked(checked)

            else:
                button.setChecked(False)

    @Slot(int, bool)
    def _on_menu_click(self, index: int) -> None:
        self._on_click(index, True)



import math

import pytest

from pyside_app_core.services.serial_service.conversion_utils import (
    decode_float_list,
    encode_float_list,
)
from pyside_app_core.types.numeric import FloatPrecision


@pytest.mark.parametrize(
    ("floats", "precision", "expected"),
    [
        ([1.1, 1.2, 1.3], "single", b"\x03\x00\xcd\xcc\x8c?\x9a\x99\x99?ff\xa6?"),
        ([10, 12.2345, 100.3], "single", b"\x03\x00\x00\x00 A\x83\xc0CA\x9a\x99\xc8B"),
        (
            [10, 12.2345, 100.3],
            "double",
            b"\x03\x00\x00\x00\x00\x00\x00\x00$@\xf2\xd2Mb\x10x(@33333\x13Y@",
        ),
        (
            [math.pi, 12.2345, 100.30000031],
            "double",
            b"\x03\x00\x18-DT\xfb!\t@\xf2\xd2Mb\x10x(@Z\x0f\x8043\x13Y@",
        ),
    ],
)
def test_encode_float_list(floats: list[float], precision: FloatPrecision, expected: bytes) -> None:
    result = encode_float_list(floats, precision)

    assert bytes(result) == expected


@pytest.mark.parametrize(
    "encoded, precision, expected",
    [
        (b"\x03\x00\xcd\xcc\x8c?\x9a\x99\x99?ff\xa6?", "single", [1.1, 1.2, 1.3]),
        (b"\x03\x00\x00\x00 A\x83\xc0CA\x9a\x99\xc8B", "single", [10, 12.2345, 100.3]),
        (
            b"\x03\x00\x00\x00\x00\x00\x00\x00$@\xf2\xd2Mb\x10x(@33333\x13Y@",
            "double",
            [10, 12.2345, 100.3],
        ),
        (
            b"\x03\x00PERT\xfb!\t@\xf2\xd2Mb\x10x(@Z\x0f\x8043\x13Y@",
            "double",
            [math.pi, 12.2345, 100.30000031],
        ),
    ],
)
def test_decode_float_list(encoded: bytes, precision: FloatPrecision, expected: list[float]) -> None:
    result = decode_float_list(encoded, precision)

    assert len(result) == len(expected)

    for original, decoded in zip(expected, result, strict=False):
        assert original == pytest.approx(decoded)



__version__ = "1.0.0"

import sys
from pathlib import Path

__lib = Path(__file__).parent.parent.parent / "src"
__root = Path(__file__).parent.parent

# these path modifications are only here to make local development easy
# when using `pyside-app-core` as a library this is not needed
sys.path.extend([str(__lib), str(__root)])



from typing import Literal

from pyside_app_core.types.numeric import FloatPrecision

# ------------------------------------------------------------------------------
# defined values
DATA_ENCODING_ENDIAN: Literal["little"] = "little"
LIST_DATA_LEN_BYTES = 2
FLOAT_PRECISION: FloatPrecision = "single"

# ------------------------------------------------------------------------------
# calculated values
DATA_STRUCT_ENDIAN = "<" if DATA_ENCODING_ENDIAN == "little" else ">"
STRUCT_FLOAT_FMT = "f" if FLOAT_PRECISION == "single" else "d"

LIST_DATA_LEN_BITS = 8 * LIST_DATA_LEN_BYTES
LIST_DATA_LEN_MAX = pow(2, LIST_DATA_LEN_BITS) - 1

# ------------------------------------------------------------------------------
COBS_SEP = b"\x00"
BUTTON_HEIGHT = 30



import pytest

from pyside_app_core.utils import strings


@pytest.mark.parametrize(
    ("width", "padded", "text", "expected"),
    [
        (80, False, "", ""),
        (10, False, "12345", "12345"),
        (10, True, "12345", "12345     "),
        (5, False, "12345678910", "12345\n67891\n0"),
        (5, False, "12345\n678910", "12345\n67891\n0"),
        (5, False, "12345678\n910", "12345\n678\n910"),
        (5, True, "12345678\n910", "12345\n678  \n910  "),
    ],
)
def test_wrap_text(width: int, padded: bool, text: str, expected: str) -> None:
    result = strings.wrap_text(text, width=width, padded=padded)
    assert result == expected



from typing import cast

from PySide6.QtCore import QByteArray, QObject
from PySide6.QtWidgets import QMainWindow, QWidget

from pyside_app_core.qt.widgets.object_name_mixin import ObjectNameMixin
from pyside_app_core.qt.widgets.settings_mixin import SettingsMixin


class WindowSettingsMixin(ObjectNameMixin, SettingsMixin):
    def __init__(self, parent: QObject | None = None, *args: object, **kwargs: object) -> None:
        super().__init__(*args, parent=parent, **kwargs)

    def _store_state(self) -> None:
        self._settings.setValue(f"{self.obj_name}_geometry", cast(QWidget, self).saveGeometry())

        if isinstance(self, QMainWindow):
            cast(SettingsMixin, self)._settings.setValue(  # noqa: SLF001
                f"{cast(ObjectNameMixin, self).obj_name}_window_state", self.saveState()
            )

    def _restore_state(self) -> None:
        cast(QWidget, self).restoreGeometry(cast(QByteArray, self._settings.value(f"{self.obj_name}_geometry")))

        if isinstance(self, QMainWindow):
            self.restoreState(cast(QByteArray, self._settings.value(f"{self.obj_name}_window_state")))



import sys
import traceback
from types import TracebackType
from typing import Protocol

from PySide6.QtWidgets import QDialog

from pyside_app_core import log


class ErrorDialogInterface(Protocol):
    def __init__(self, etype: type[BaseException], message: str, details: str | None = None): ...

    def exec(self) -> int: ...


_error_dialog_class: type[ErrorDialogInterface] | None = None


def install_excepthook(error_dialog: type[QDialog] | None) -> None:
    global _error_dialog_class  # noqa
    _error_dialog_class = error_dialog
    sys.excepthook = __custom_excepthook


def __custom_excepthook(
    etype: type[BaseException],
    evalue: BaseException,
    tb: TracebackType | None,
) -> None:
    formatted_exception_string = "".join(traceback.format_exception(etype, evalue, tb))

    log.exception(formatted_exception_string)

    if hasattr(evalue, "internal") and evalue.internal:
        return

    if _error_dialog_class:
        _error_dialog_class(etype, str(evalue), formatted_exception_string).exec()



import pytest

from pyside_app_core.services.serial_service.float_map import FloatMap


def test_float_map__init() -> None:
    fm = FloatMap[int](
        {
            1: 1.0,
            2: 2.0,
            3: 3.0,
        }
    )

    assert fm[1] == 1.0
    assert fm[2] == 2.0
    assert fm[3] == 3.0
    assert len(fm) == 3

    with pytest.raises(KeyError):
        _ = fm[99]

    with pytest.raises(ValueError) as ee:
        FloatMap({})
    assert str(ee.value) == "can't create an empty FloatMap"

    with pytest.raises(ValueError) as ve:
        FloatMap(dict.fromkeys(range(65536), 1.0))
    assert str(ve.value) == "data has too many values, must fit in 2 bytes"

    with pytest.raises(ValueError) as ke:
        FloatMap(dict.fromkeys(range(60000, 70000), 1.0))
    assert str(ke.value) == 'key: "65536" does not fit in 2 bytes'

    with pytest.raises(ValueError) as se:
        FloatMap({-1: 1.0})
    assert str(se.value) == 'key: "-1" is not a positive number'

    for k, v in fm.items():
        assert k == v  # incidental, just want to be able to assert in loop


@pytest.mark.parametrize(
    "count, expected",
    [
        (1, "<HHf"),
        (2, "<HHfHf"),
        (3, "<HHfHfHf"),
        (0, None),
    ],
)
def test_float_map__pack_format(count: int, expected: str | None) -> None:
    if count < 1:
        with pytest.raises(ValueError):
            FloatMap.pack_format(count)
    else:
        result = FloatMap.pack_format(count)
        assert result == expected


@pytest.mark.parametrize(
    "data, expected",
    [
        ({1: 1.0}, b"\x01\x00\x01\x00\x00\x00\x80?"),
        ({1: 1.0, 2: 2.0}, b"\x02\x00\x01\x00\x00\x00\x80?\x02\x00\x00\x00\x00@"),
        (
            {1: 1.0, 2: 2.0, 3: 3.0},
            b"\x03\x00\x01\x00\x00\x00\x80?\x02\x00\x00\x00\x00@\x03\x00\x00\x00@@",
        ),
    ],
)
def test_float_map__pack(data: dict[int, float], expected: bytes) -> None:
    result = FloatMap[int](data).pack()
    assert result == expected


@pytest.mark.parametrize(
    "raw, expected",
    [
        (b"\x01\x00\x01\x00\x00\x00\x80?", {1: 1.0}),
        (b"\x02\x00\x01\x00\x00\x00\x80?\x02\x00\x00\x00\x00@", {1: 1.0, 2: 2.0}),
        (
            b"\x03\x00\x01\x00\x00\x00\x80?\x02\x00\x00\x00\x00@\x03\x00\x00\x00@@",
            {1: 1.0, 2: 2.0, 3: 3.0},
        ),
    ],
)
def test_float_map__unpack(raw: bytes, expected: dict[int, float]) -> None:
    result = FloatMap[int].unpack(raw)
    assert result == expected



import pytest

from pyside_app_core.utils import compare


@pytest.mark.parametrize(
    ("a", "b", "expected"),
    [
        (1, 1, True),
        (1.0, 1.0, True),
        (1.000001, 1.000001, True),
        (1.0, 1.000001, True),
        (1.0, 1.00001, False),
        (1.0, 1.1, False),
        (100000000.0, 100000000.0, True),
        (100000000.0, 100000000.0000001, True),
        (100000000.0, 100000000.00001, False),
    ],
)
def test_utils_compare_float_approx(a: float, b: float, expected: bool) -> None:
    result = compare.float_approx(a, b)
    assert result == expected



from typing import ClassVar, Generic, NamedTuple, TypeVar

from pyside_app_core.errors.basic_errors import ApplicationError

_FVT = TypeVar("_FVT")


class Field(Generic[_FVT]):
    def __init__(self, _: type[_FVT]) -> None:
        self.value: _FVT | None = None

    def __set_name__(self, owner: type, name: str) -> None:
        owner._fields.append(name)  # type: ignore[attr-defined]

    def __get__(self, obj: object, objtype: type) -> _FVT:
        if self.value is None:
            raise ValueError("Field is not initialized")
        return self.value

    def __set__(self, obj: object, value: _FVT) -> None:
        self.value = value


class NullableField(Field[_FVT | None]):
    def __get__(self, obj: object, objtype: type) -> _FVT | None:
        return self.value

    def __set__(self, obj: object, value: _FVT | None) -> None:
        self.value = value


class MetadataMeta(type):
    _initialized = False
    _fields: ClassVar[list[str]] = []

    def __str__(cls) -> str:
        if not cls._initialized:
            raise ApplicationError("Meta not initialized")

        vals: list[str] = []

        for f in cls._fields:
            val = cls.__dict__.get(f)
            vals.append(f"{f}: {val}")

        return f"<{cls.__name__}: [{', '.join(vals)}]>"


class TemplateMeta(NamedTuple):
    resource: str
    data: dict[str, str] | None = None


class AppMetadata(metaclass=MetadataMeta):
    id = Field(str)
    name = Field(str)
    version = Field(str)
    icon = Field(str)
    about_template = NullableField(TemplateMeta)
    help_url = Field(str)
    bug_report_url = Field(str)
    oss_licenses = Field(list)

    @classmethod
    def init(
        cls,
        app_id: str,
        name: str,
        version: str,
        icon_resource: str = "",
        about_template: TemplateMeta | None = None,
        help_url: str = "",
        bug_report_url: str = "",
        oss_licenses: list[str] | None = None,
    ) -> None:
        if cls._initialized:
            raise ApplicationError("Metadata already initialized")

        cls.id = app_id  # type: ignore[assignment]
        cls.name = name  # type: ignore[assignment]
        cls.version = version  # type: ignore[assignment]
        cls.icon = icon_resource  # type: ignore[assignment]
        cls.about_template = about_template  # type: ignore[assignment]
        cls.help_url = help_url  # type: ignore[assignment]
        cls.bug_report_url = bug_report_url  # type: ignore[assignment]
        cls.oss_licenses = [  # type: ignore[assignment]
            ":/core/notices/licenses/python.md",
            ":/core/notices/licenses/qt.md",
            ":/core/notices/licenses/jinja2.md",
            ":/core/notices/licenses/iconoir.md",
            *(oss_licenses or []),
        ]

        cls._initialized = True



from PySide6 import QtSerialPort

from pyside_app_core.errors.basic_errors import CoreError


class SerialError(CoreError):
    def __init__(
        self,
        port: QtSerialPort.QSerialPort | None,
        error: QtSerialPort.QSerialPort.SerialPortError,
        *,
        internal: bool = False,
    ):
        if port:
            msg = f"Serial error{f' {error} ' if error else ' '}on port: {port.portName()}"
        else:
            msg = "Serial port not configured"
        super().__init__(msg, internal=internal)


class SerialUnknownError(SerialError):
    """something went wrong"""


class SerialConnectionError(SerialError):
    """a connection error occurred"""


class SerialReadError(SerialError):
    """a data read error occurred"""


class SerialWriteError(SerialError):
    """a data write error occurred"""


class SerialDisconnectedError(SerialError):
    """connection was broken"""



from PySide6 import QtWidgets

from pyside_app_core.utils import strings


class ErrorDialog(QtWidgets.QMessageBox):
    CHAR_WIDTH = 120

    def __init__(
        self,
        etype: type[BaseException],
        msg: str,
        details: str,
        parent: QtWidgets.QWidget | None = None,
    ):
        super().__init__(parent=parent)
        self.setWindowTitle("Error")

        self.setText(f"Error: {etype.__name__: <{self.CHAR_WIDTH}}")
        if msg:
            self.setInformativeText(strings.wrap_text(msg, width=self.CHAR_WIDTH, padded=True) + "\n")
        if details:
            self.setDetailedText(details)



import struct

from pyside_app_core.constants import (
    DATA_ENCODING_ENDIAN,
    DATA_STRUCT_ENDIAN,
    FLOAT_PRECISION,
    LIST_DATA_LEN_BYTES,
    LIST_DATA_LEN_MAX,
)
from pyside_app_core.errors.encode_errors import EncodingListError
from pyside_app_core.types.numeric import FloatPrecision


def int_to_bytes(val: int, *, num_bytes: int, signed: bool) -> bytes:
    return val.to_bytes(length=num_bytes, byteorder=DATA_ENCODING_ENDIAN, signed=signed)


def int_from_bytes(val: bytes, *, signed: bool) -> int:
    return int.from_bytes(bytes=val, byteorder=DATA_ENCODING_ENDIAN, signed=signed)


def encode_float_list(floats: list[float], precision: FloatPrecision = FLOAT_PRECISION) -> bytearray:
    data_len = len(floats)
    if data_len > LIST_DATA_LEN_MAX:
        raise EncodingListError(data_len)

    precision_format = "f" if precision == "single" else "d"

    encoded = bytearray()
    encoded.extend(int_to_bytes(data_len, num_bytes=LIST_DATA_LEN_BYTES, signed=False))
    encoded.extend(struct.pack(f"{DATA_STRUCT_ENDIAN}{data_len}{precision_format}", *floats))

    return encoded


def decode_float_list(raw_data: bytes, precision: FloatPrecision = FLOAT_PRECISION) -> list[float]:
    len_bytes = raw_data[:LIST_DATA_LEN_BYTES]
    data_bytes = raw_data[LIST_DATA_LEN_BYTES:]

    precision_format = "f" if precision == "single" else "d"

    data_len = int_from_bytes(len_bytes, signed=False)
    return list(struct.unpack(f"{DATA_STRUCT_ENDIAN}{data_len}{precision_format}", data_bytes))



import dataclasses
from pathlib import Path


@dataclasses.dataclass(frozen=True)
class QtResourceFile:
    path: str | Path
    alias: str | None


@dataclasses.dataclass(frozen=True)
class QtResourceGroup:
    prefix: str | None
    lang: str | None
    files: list[QtResourceFile]



from typing import cast

from PySide6.QtCore import Qt
from PySide6.QtWidgets import QWidget


class ObjectNameMixin:
    def __init__(self, *args: object, **kwargs: object) -> None:
        super().__init__(*args, **kwargs)

        if isinstance(self, QWidget):
            self.setAttribute(Qt.WidgetAttribute.WA_StyledBackground)

        obj_name = self.__class__.__name__
        if hasattr(self, "OBJECT_NAME"):
            obj_name = self.OBJECT_NAME.replace(" ", "_")

        cast(QWidget, self).setObjectName(obj_name)

    @property
    def obj_name(self) -> str:
        return cast(QWidget, self).objectName()



import os
import platform

__valid_platforms = ["Darwin", "Windows", "Linux"]


def _platform() -> str:
    system_platform = platform.system()
    override = os.environ.get("PLATFORM_OVERRIDE", system_platform)

    return override if override in __valid_platforms else system_platform


__platform: str = _platform()

is_macos = __platform == "Darwin"
is_windows = __platform == "Windows"
is_linux = __platform == "Linux"



from PySide6.QtCore import QFile, QTextStream


def load_text_file(file_url: str) -> str:
    qfile = QFile(file_url)
    qfile.open(QFile.OpenModeFlag.ReadOnly | QFile.OpenModeFlag.Text)
    ts = QTextStream(qfile)
    return ts.readAll()



from typing import cast

from PySide6.QtCore import QIODevice, QObject, QTimer, Signal
from PySide6.QtSerialPort import QSerialPort, QSerialPortInfo

from pyside_app_core import log
from pyside_app_core.errors.serial_errors import (
    SerialConnectionError,
    SerialDisconnectedError,
    SerialError,
    SerialReadError,
    SerialUnknownError,
    SerialWriteError,
)
from pyside_app_core.services.serial_service.types import (
    Encodable,
    PortFilter,
    SerialReader,
    TranscoderInterface,
)
from pyside_app_core.utils.time_ms import SECONDS


def _noop(ports: list[QSerialPortInfo]) -> list[QSerialPortInfo]:
    log.debug(f"Filtering ports: {ports}")
    return ports


class SerialService(QObject):
    ports = Signal(list)
    com_connect = Signal(QSerialPort)
    com_disconnect = Signal()
    data = Signal(object)
    error = Signal(Exception)

    def __init__(self, transcoder: type[TranscoderInterface], parent: QObject):
        super().__init__(parent=parent)

        self._port_filter: PortFilter = _noop

        self._transcoder: type[TranscoderInterface] = transcoder
        self._com: QSerialPort | None = None
        self._buffer = bytearray()

    def _new_com(self, port_info: QSerialPortInfo) -> tuple[QSerialPort, QSerialPort.SerialPortError | None]:
        com = QSerialPort(port_info, parent=self)
        com.setBaudRate(QSerialPort.BaudRate.Baud115200)
        open_ok = com.open(QIODevice.OpenModeFlag.ReadWrite)

        return com, None if open_ok else com.error()

    def set_port_filter(self, func: PortFilter) -> None:
        log.debug(f"updating port filter: {func}")
        self._port_filter = func

    def open_connection(self, port_info: QSerialPortInfo) -> bool:
        self.close_connection()

        self._com, error = self._new_com(port_info)
        if error:
            self._on_error(error)
            return False

        self._com.readyRead.connect(self._on_data)
        self._com.errorOccurred.connect(self._on_error)

        self.com_connect.emit(self._com)

        return True

    def register_reader(self, reader: SerialReader) -> None:
        log.debug(f"Registering reader {reader}")
        self.com_connect.connect(reader.handle_serial_connect)
        self.com_disconnect.connect(reader.handle_serial_disconnect)
        self.ports.connect(reader.handle_serial_ports)
        self.data.connect(reader.handle_serial_data)
        self.error.connect(reader.handle_serial_error)

    def scan_for_ports(self) -> None:
        log.debug("Scanning for ports...")
        ports = QSerialPortInfo.availablePorts()

        self._debug_ports(ports)

        filtered_ports = self._port_filter(ports)
        log.debug(f"Sending ports: {filtered_ports}")

        self.ports.emit(filtered_ports)

    def send_data(self, data: Encodable) -> None:
        log.debug(f"sending data: {data}")

        if not self._com:
            log.debug("com port not connected")
            return

        self._com.write(self._transcoder.encode(data))

    def close_connection(self) -> None:
        if not self._com:
            return

        self.com_disconnect.emit()

        try:
            self._com.errorOccurred.disconnect()
            self._com.readyRead.disconnect()
        except Exception as e:  # noqa: BLE001
            log.exception(e)

        if self._com and self._com.isOpen():
            self._com.flush()
            self._com.close()

        self._com.deleteLater()
        self._com = None

    def deleteLater(self) -> None:  # noqa: N802
        self.close_connection()
        super().deleteLater()

    def _on_data(self, *_: object, **__: object) -> None:
        if not self._com:
            return

        self._buffer.extend(cast(bytearray, self._com.readAll()))

        chunks, remainder = self._transcoder.process_buffer(self._buffer)

        for chunk in chunks:
            try:
                result = self._transcoder.decode(chunk)
                log.debug(f"received data: {result}")
                self.data.emit(result)
            except Exception as e:  # noqa: BLE001
                log.exception(e)
                self.error.emit(e)

            self._buffer.clear()
            if remainder is not None:
                self._buffer.extend(remainder)

    def _on_error(self, error: QSerialPort.SerialPortError | None) -> None:
        exception: SerialError

        if error is None or error == QSerialPort.SerialPortError.NoError:
            return
        if error == QSerialPort.SerialPortError.OpenError:
            exception = SerialConnectionError(self._com, error)
        elif error == QSerialPort.SerialPortError.ReadError:
            exception = SerialReadError(self._com, error)
        elif error == QSerialPort.SerialPortError.WriteError:
            exception = SerialWriteError(self._com, error)
        elif error == QSerialPort.SerialPortError.ResourceError:
            exception = SerialDisconnectedError(self._com, error, internal=True)
        else:
            exception = SerialUnknownError(self._com, error)

        self.close_connection()

        QTimer.singleShot(3 * SECONDS, self.scan_for_ports)

        self.error.emit(exception)
        raise exception

    def _debug_ports(self, ports: list[QSerialPortInfo]) -> None:
        for p in ports:
            log.debug("-----------------------------")
            log.debug(f"name:         {p.portName()}")
            log.debug(f"manufacturer: {p.manufacturer()}")
            log.debug(f"productId:    {p.productIdentifier()}")
            log.debug(f"serialNumber: {p.serialNumber()}")
            log.debug(f"vendorId:     {p.vendorIdentifier()}")
            log.debug(f"systemLoc:    {p.systemLocation()}")



from typing import cast

from jinja2 import BaseLoader, Environment
from PySide6.QtCore import Qt
from PySide6.QtGui import QPalette
from PySide6.QtWidgets import (
    QApplication,
    QDialog,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QTabWidget,
    QTextBrowser,
    QVBoxLayout,
    QWidget,
)

from pyside_app_core.qt.application_service import AppMetadata, TemplateMeta
from pyside_app_core.utils.files import load_text_file


class AboutDialog(QDialog):
    def __init__(self) -> None:
        super().__init__()

        self.setFixedSize(640, 480)
        self.setWindowTitle(f"About {AppMetadata.name}")

        _ly = QVBoxLayout()
        _ly.setContentsMargins(0, 0, 0, 10)
        _ly.setSpacing(5)
        self.setLayout(_ly)

        _tabs = QTabWidget(parent=self)
        _ly.addWidget(_tabs)

        _tabs.addTab(_AboutApp(parent=self), "About")
        _tabs.addTab(_Licenses(parent=self), "Licenses")

        _ly_btns = QHBoxLayout()
        _ly_btns.setContentsMargins(10, 0, 10, 0)
        _close = QPushButton("Close", self)
        _close.setDefault(True)
        _ly_btns.addWidget(_close, alignment=Qt.AlignmentFlag.AlignRight)
        _ly.addLayout(_ly_btns)

        _close.clicked.connect(self.close)


class _AboutApp(QWidget):
    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent=parent)
        _ly_split = QHBoxLayout()
        _ly_split.setContentsMargins(15, 1, 1, 2)
        _ly_split.setSpacing(10)
        self.setLayout(_ly_split)

        _ly_icon = QVBoxLayout()
        _ly_icon.setContentsMargins(0, 10, 0, 0)
        icon = QLabel(parent=self)
        icon.setPixmap(cast(QApplication, QApplication.instance()).windowIcon().pixmap(64, 64))
        _ly_icon.addWidget(icon, stretch=1, alignment=Qt.AlignmentFlag.AlignTop)
        _ly_split.addLayout(_ly_icon)

        text_area = QTextBrowser()
        text_area.setViewportMargins(10, 10, 10, 10)
        text_area.setStyleSheet(
            f"""
        QTextBrowser {{
            background-color: {self.palette().color(QPalette.ColorRole.Window).name()};
            border: none;
        }}
        """
        )
        _ly_split.addWidget(text_area, stretch=99)

        args: dict[str, str] = {
            "name": AppMetadata.name,
            "version": AppMetadata.version,
            "id": AppMetadata.id,
        }

        env = Environment(loader=BaseLoader(), autoescape=True)
        if template_data := AppMetadata.about_template:
            raw, data = cast(TemplateMeta, template_data)
            args = {**args, **(data or {})}
        else:
            raw = load_text_file(":/core/notices/about.md.jinja2")

        template = env.from_string(raw)

        text_area.setMarkdown(template.render(**args))


class _Licenses(QWidget):
    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent=parent)
        _ly = QVBoxLayout()
        _ly.setContentsMargins(1, 1, 1, 1)
        self.setLayout(_ly)

        text_area = QTextBrowser()
        text_area.setViewportMargins(10, 10, 10, 10)
        text_area.setStyleSheet(
            f"""
        QTextBrowser {{
            background-color: {self.palette().color(QPalette.ColorRole.Window).name()};
            border: none;
        }}
        """
        )
        _ly.addWidget(text_area, stretch=99)

        markdown = """
# Open Source Licenses
"""
        for lic in AppMetadata.oss_licenses:
            markdown = f"""{markdown}

---

{load_text_file(lic)}"""

        text_area.setMarkdown(markdown)



from typing import cast

from PySide6.QtWidgets import QWidget


class BaseMixin:
    def __init__(
        self,
        *args: object,
        **kwargs: object,
    ):
        super().__init__(*args, **kwargs)

        cast(QWidget, self).setMouseTracking(True)
        cast(QWidget, self).setMinimumSize(256, 128)



import contextlib
from collections.abc import Iterator

from PySide6 import QtGui
from PySide6.QtGui import QAction
from PySide6.QtWidgets import (
    QMenu,
    QMenuBar,
    QWidget,
)


class MenuContext(QMenu):
    def __init__(self, name: str, parent: QWidget):
        super().__init__(name, parent)
        self._menu_map: dict[str, MenuContext] = {}
        self._action_map: dict[str, QAction] = {}

    @contextlib.contextmanager
    def menu(self, name: str) -> Iterator["MenuContext"]:
        if name not in self._menu_map:
            self._menu_map[name] = MenuContext(name, self)
            self._menu_map[name].setObjectName(f"Menu_{name.replace(' ', '_')}")
            self.addMenu(self._menu_map[name])

        yield self._menu_map[name]

    @contextlib.contextmanager
    def action(self, name: str) -> Iterator[QAction]:
        if name not in self._action_map:
            self._action_map[name] = QtGui.QAction(name, self)
            self._action_map[name].setObjectName(f"MenuAction_{name.replace(' ', '_')}")
            self.addAction(self._action_map[name])

        yield self._action_map[name]


class MenuBarContext(QMenuBar):
    def __init__(self, parent: QWidget):
        super().__init__(parent=parent)

        self._menu_map: dict[str, MenuContext] = {}

    @contextlib.contextmanager
    def menu(self, name: str) -> Iterator[MenuContext]:
        if name not in self._menu_map:
            self._menu_map[name] = MenuContext(name, self)
            self._menu_map[name].setObjectName(f"Menu_{name.replace(' ', '_')}")
            self.addMenu(self._menu_map[name])

        yield self._menu_map[name]



from pathlib import Path

from pyside_app_core.qt.widgets.base_app import BaseApp
from toolbar_app.main_window import SimpleMainWindow


class SimpleApp(BaseApp):
    def __init__(self, resources_rcc: Path | None = None):
        super().__init__(resources_rcc=resources_rcc)

    def build_main_window(self):
        return SimpleMainWindow()
